/*
 * DemosceneMusic - DataManager.java - Copyright © 2012 David Roden
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package net.pterodactylus.demoscenemusic.data;

import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Collection;
import java.util.EnumMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.UUID;
import java.util.concurrent.Callable;

import net.pterodactylus.demoscenemusic.data.Track.Relationship;
import net.pterodactylus.util.collection.Memoizer;
import net.pterodactylus.util.database.Database;
import net.pterodactylus.util.database.DatabaseException;
import net.pterodactylus.util.database.Field;
import net.pterodactylus.util.database.Join;
import net.pterodactylus.util.database.Join.JoinType;
import net.pterodactylus.util.database.ObjectCreator;
import net.pterodactylus.util.database.ObjectCreator.StringCreator;
import net.pterodactylus.util.database.OrderField;
import net.pterodactylus.util.database.Parameter.IntegerParameter;
import net.pterodactylus.util.database.Parameter.StringParameter;
import net.pterodactylus.util.database.Query;
import net.pterodactylus.util.database.Query.Type;
import net.pterodactylus.util.database.ResultProcessor;
import net.pterodactylus.util.database.ValueField;
import net.pterodactylus.util.database.ValueFieldWhereClause;

/**
 * Interface between the database and the application.
 *
 * @author <a href="mailto:bombe@pterodactylus.net">David ‘Bombe’ Roden</a>
 */
public class DataManager {

	/** The artist object creator. */
	@SuppressWarnings("synthetic-access")
	private final ObjectCreator<Artist> artistCreator = new ArtistCreator();

	/** The group object creator. */
	@SuppressWarnings("synthetic-access")
	private final ObjectCreator<Group> groupCreator = new GroupCreator();

	/** The track object creator. */
	@SuppressWarnings("synthetic-access")
	private final ObjectCreator<Track> trackCreator = new TrackCreator();

	/** The track object creator. */
	@SuppressWarnings("synthetic-access")
	private final ObjectCreator<TrackDerivative> trackDerivativeCreator = new TrackDerivativeCreator();

	/** The style object creator. */
	@SuppressWarnings("synthetic-access")
	private final ObjectCreator<Style> styleCreator = new StyleCreator();

	/** The party object creator. */
	@SuppressWarnings("synthetic-access")
	private final ObjectCreator<Party> partyCreator = new PartyCreator();

	/** The {@link User} object creator. */
	@SuppressWarnings("synthetic-access")
	private final ObjectCreator<User> userCreator = new UserCreator();

	/** The database. */
	private final Database database;

	/**
	 * Creates a new data manager.
	 *
	 * @param database
	 *            The database to operate on
	 */
	public DataManager(Database database) {
		this.database = database;
	}

	/**
	 * Returns all artists.
	 *
	 * @return All artists
	 * @throws DatabaseException
	 *             if a database error occurs
	 */
	public Collection<Artist> getAllArtists() throws DatabaseException {
		Query query = new Query(Type.SELECT, "ARTISTS");
		query.addField(new Field("ARTISTS.*"));
		return loadArtistProperties(database.getMultiple(query, artistCreator));
	}

	/**
	 * Returns the artist with the given ID.
	 *
	 * @param id
	 *            The ID of the artist
	 * @return The artist with the given ID, or {@code null} if there is no
	 *         artist with the given ID
	 * @throws DatabaseException
	 *             if a database error occurs
	 */
	public Artist getArtistById(String id) throws DatabaseException {
		Query query = new Query(Type.SELECT, "ARTISTS");
		query.addField(new Field("ARTISTS.*"));
		query.addWhereClause(new ValueFieldWhereClause(new ValueField("ARTISTS.ID", new StringParameter(id))));
		return loadArtistProperties(database.getSingle(query, artistCreator));
	}

	/**
	 * Returns all artists that belong to the group with the given ID.
	 *
	 * @param groupId
	 *            The ID of the group
	 * @return All artists belonging to the given group
	 * @throws DatabaseException
	 *             if a database error occurs
	 */
	public Collection<Artist> getArtistsByGroup(String groupId) throws DatabaseException {
		Query query = new Query(Type.SELECT, "ARTISTS");
		query.addField(new Field("ARTISTS.*"));
		query.addJoin(new Join(JoinType.INNER, "GROUP_ARTISTS", new Field("ARTISTS.ID"), new Field("GROUP_ARTISTS.ARTIST")));
		query.addWhereClause(new ValueFieldWhereClause(new ValueField("GROUP_ARTISTS.GROUP_", new StringParameter(groupId))));
		return loadArtistProperties(database.getMultiple(query, artistCreator));
	}

	/**
	 * Returns all artists involved in the track with the given ID.
	 *
	 * @param trackId
	 *            The ID of the track
	 * @return All artists involved in the track, in preferred order
	 * @throws DatabaseException
	 *             if a database error occurs
	 */
	public List<Artist> getArtistsByTrack(String trackId) throws DatabaseException {
		Query query = new Query(Type.SELECT, "ARTISTS");
		query.addField(new Field("ARTISTS.*"));
		query.addJoin(new Join(JoinType.INNER, "TRACK_ARTISTS", new Field("TRACK_ARTISTS.ARTIST"), new Field("ARTISTS.ID")));
		query.addWhereClause(new ValueFieldWhereClause(new ValueField("TRACK_ARTISTS.TRACK", new StringParameter(trackId))));
		query.addOrderField(new OrderField(new Field("TRACK_ARTISTS.DISPLAY_ORDER")));
		return loadArtistProperties(database.getMultiple(query, artistCreator));
	}

	/**
	 * Creates a new artist with the given name.
	 *
	 * @param name
	 *            The name of the artist
	 * @return The created artist
	 * @throws DatabaseException
	 *             if a database error occurs
	 */
	public Artist createArtist(String name) throws DatabaseException {
		Query query = new Query(Type.INSERT, "ARTISTS");
		String id = UUID.randomUUID().toString();
		query.addValueField(new ValueField("ID", new StringParameter(id)));
		query.addValueField(new ValueField("NAME", new StringParameter(name)));
		database.insert(query);
		return loadArtistProperties(getArtistById(id));
	}

	/**
	 * Saves the given artist.
	 *
	 * @param artist
	 *            The artist to save
	 * @throws DatabaseException
	 *             if a database error occurs
	 */
	public void saveArtist(Artist artist) throws DatabaseException {
		Query query = new Query(Type.UPDATE, "ARTISTS");
		query.addValueField(new ValueField("NAME", new StringParameter(artist.getName())));
		query.addWhereClause(new ValueFieldWhereClause(new ValueField("ID", new StringParameter(artist.getId()))));
		database.update(query);

		/* save groups. */
		Collection<Group> groups = artist.getGroups();
		query = new Query(Type.DELETE, "GROUP_ARTISTS");
		query.addWhereClause(new ValueFieldWhereClause(new ValueField("GROUP_ARTISTS.ARTIST", new StringParameter(artist.getId()))));
		database.update(query);
		for (Group group : groups) {
			query = new Query(Type.INSERT, "GROUP_ARTISTS");
			query.addValueField(new ValueField("GROUP_", new StringParameter(group.getId())));
			query.addValueField(new ValueField("ARTIST", new StringParameter(artist.getId())));
			database.insert(query);
		}

		/* save properties. */
		saveArtistProperties(artist);
	}

	/**
	 * Saves the properties of the given artist.
	 *
	 * @param artist
	 *            The artist whose properties to save
	 * @throws DatabaseException
	 *             if a database error occurs
	 */
	public void saveArtistProperties(Artist artist) throws DatabaseException {
		saveProperties(artist.getProperties(), "ARTIST_PROPERTIES", "ARTIST", artist.getId());
	}

	/**
	 * Saves the given properties to the given table for the given principal.
	 *
	 * @param properties
	 *            The properties to save
	 * @param table
	 *            The table in which to save the properties
	 * @param type
	 *            The type of the principal (e. g. “ARTIST” or “TRACK”)
	 * @param id
	 *            The ID of the principial
	 * @throws DatabaseException
	 *             if a database error occurs
	 */
	public void saveProperties(Properties properties, String table, String type, String id) throws DatabaseException {
		if (!properties.isDirty()) {
			return;
		}
		Query query = new Query(Type.DELETE, table);
		query.addWhereClause(new ValueFieldWhereClause(new ValueField(type, new StringParameter(id))));
		database.update(query);
		for (Entry<String, String> property : properties) {
			Query insertQuery = new Query(Type.INSERT, table);
			insertQuery.addValueField(new ValueField(type, new StringParameter(id)));
			insertQuery.addValueField(new ValueField("PROPERTY", new StringParameter(property.getKey())));
			insertQuery.addValueField(new ValueField("VALUE", new StringParameter(property.getValue())));
			database.insert(insertQuery);
		}
		properties.resetDirty();
	}

	/**
	 * Loads the properties for an artist.
	 *
	 * @param artist
	 *            The artist to load the properties for
	 * @return The artist
	 * @throws DatabaseException
	 *             if a database error occurs
	 */
	public Artist loadArtistProperties(final Artist artist) throws DatabaseException {
		return loadProperties(artist, "ARTIST_PROPERTIES", "ARTIST");
	}

	/**
	 * Loads the properties of all given artists.
	 *
	 * @param artists
	 *            The artists to load the properties for
	 * @return The list of artists
	 * @throws DatabaseException
	 *             if a database error occurs
	 */
	public List<Artist> loadArtistProperties(List<Artist> artists) throws DatabaseException {
		for (Artist artist : artists) {
			loadArtistProperties(artist);
		}
		return artists;
	}

	/**
	 * Returns all remix artists involved in the track with the given ID.
	 *
	 * @param trackId
	 *            The ID of the track
	 * @return All remix artists involved in the track, in preferred order
	 * @throws DatabaseException
	 *             if a database error occurs
	 */
	public List<Artist> getRemixArtistsByTrack(String trackId) throws DatabaseException {
		Query query = new Query(Type.SELECT, "ARTISTS");
		query.addField(new Field("ARTISTS.*"));
		query.addJoin(new Join(JoinType.INNER, "TRACK_REMIX_ARTISTS", new Field("TRACK_REMIX_ARTISTS.ARTIST"), new Field("ARTISTS.ID")));
		query.addWhereClause(new ValueFieldWhereClause(new ValueField("TRACK_REMIX_ARTISTS.TRACK", new StringParameter(trackId))));
		query.addOrderField(new OrderField(new Field("TRACK_REMIX_ARTISTS.DISPLAY_ORDER")));
		return database.getMultiple(query, artistCreator);
	}

	/**
	 * Returns all related tracks for the track with the given ID.
	 *
	 * @param trackId
	 *            The ID of the tracks
	 * @return A mapping from relationship to all tracks that match the relation
	 * @throws DatabaseException
	 *             if a database error occurs
	 */
	public Map<Relationship, Collection<Track>> getRelatedTracksByTrack(String trackId) throws DatabaseException {
		Query query = new Query(Type.SELECT, "TRACKS");
		query.addField(new Field("TRACKS.*"));
		query.addField(new Field("TRACK_RELATIONS.*"));
		query.addJoin(new Join(JoinType.INNER, "TRACK_RELATIONS", new Field("TRACK_RELATIONS.TRACK"), new Field("TRACK_RELATIONS.RELATED_TRACK")));
		query.addWhereClause(new ValueFieldWhereClause(new ValueField("TRACK_RELATIONS.TRACK", new StringParameter(trackId))));
		final Map<Relationship, Collection<Track>> relatedTracks = new EnumMap<Relationship, Collection<Track>>(Relationship.class);
		database.process(query, new ResultProcessor() {

			@Override
			@SuppressWarnings("synthetic-access")
			public void processResult(ResultSet resultSet) throws SQLException {
				Track track = trackCreator.createObject(resultSet);
				Relationship relationship = Relationship.valueOf(resultSet.getString("TRACK_RELATIONS.RELATIONSHIP"));
				if (!relatedTracks.containsKey(relationship)) {
					relatedTracks.put(relationship, new HashSet<Track>());
				}
				relatedTracks.get(relationship).add(track);
			}
		});
		return relatedTracks;
	}

	/**
	 * Returns the track with the given ID.
	 *
	 * @param id
	 *            The ID of the track
	 * @return The track with the given ID, or {@code null} if there is no such
	 *         track
	 * @throws DatabaseException
	 *             if a database error occurs
	 */
	public Track getTrackById(String id) throws DatabaseException {
		Query query = new Query(Type.SELECT, "TRACKS");
		query.addField(new Field("TRACKS.*"));
		query.addWhereClause(new ValueFieldWhereClause(new ValueField("TRACKS.ID", new StringParameter(id))));
		return loadTrackProperties(database.getSingle(query, trackCreator));
	}

	/**
	 * Returns the track that contains the derivative with the given ID.
	 *
	 * @param id
	 *            The ID of the track derivative
	 * @return The track the derivative belongs to, or {@code null} if there is
	 *         no such track
	 * @throws DatabaseException
	 *             if a database error occurs
	 */
	public Track getTrackByDerivativeId(String id) throws DatabaseException {
		Query query = new Query(Type.SELECT, "TRACKS");
		query.addField(new Field("TRACKS.*"));
		query.addJoin(new Join(JoinType.INNER, "TRACK_DERIVATIVES", new Field("TRACK_DERIVATIVES.TRACK"), new Field("TRACKS.ID")));
		query.addWhereClause(new ValueFieldWhereClause(new ValueField("TRACK_DERIVATIVES.ID", new StringParameter(id))));
		return loadTrackProperties(database.getSingle(query, trackCreator));
	}

	/**
	 * Returns all tracks by the artist with the given ID.
	 *
	 * @param artistId
	 *            The ID of the artist
	 * @return All tracks by the given artist
	 * @throws DatabaseException
	 *             if a database error occurs
	 */
	public Collection<Track> getTracksByArtist(String artistId) throws DatabaseException {
		Query query = new Query(Type.SELECT, "TRACKS");
		query.addField(new Field("TRACKS.*"));
		query.addJoin(new Join(JoinType.INNER, "TRACK_ARTISTS", new Field("TRACKS.ID"), new Field("TRACK_ARTISTS.TRACK")));
		query.addWhereClause(new ValueFieldWhereClause(new ValueField("TRACK_ARTISTS.ARTIST", new StringParameter(artistId))));
		return loadTrackProperties(database.getMultiple(query, trackCreator));
	}

	/**
	 * Returns all tracks that were released at the party with the given ID.
	 *
	 * @param partyId
	 *            The ID of the party
	 * @return All tracks that were released at the given party
	 * @throws DatabaseException
	 *             if a database error occurs
	 */
	public Collection<Track> getTracksByParty(String partyId) throws DatabaseException {
		Query query = new Query(Type.SELECT, "TRACKS");
		query.addField(new Field("TRACKS.*"));
		query.addJoin(new Join(JoinType.INNER, "PARTY_TRACKS", new Field("PARTY_TRACKS.TRACK"), new Field("TRACKS.ID")));
		query.addWhereClause(new ValueFieldWhereClause(new ValueField("PARTY_TRACKS.PARTY", new StringParameter(partyId))));
		return loadTrackProperties(database.getMultiple(query, trackCreator));
	}

	/**
	 * Loads the properties for the given track.
	 *
	 * @param track
	 *            The track for which to load the properties
	 * @return The track with the properties loaded
	 * @throws DatabaseException
	 *             if a database error occurs
	 */
	public Track loadTrackProperties(Track track) throws DatabaseException {
		return loadProperties(track, "TRACK_PROPERTIES", "TRACK");
	}

	/**
	 * Loads the properties for the given tracks.
	 *
	 * @param tracks
	 *            The tracks for which to load the properties
	 * @return The tracks with the properties loaded
	 * @throws DatabaseException
	 *             if a database error occurs
	 */
	public List<Track> loadTrackProperties(List<Track> tracks) throws DatabaseException {
		for (Track track : tracks) {
			loadTrackProperties(track);
		}
		return tracks;
	}

	/**
	 * Creates a new track with the given name.
	 *
	 * @param name
	 *            The name of the track
	 * @return The created track
	 * @throws DatabaseException
	 *             if a database error occurs
	 */
	public Track createTrack(String name) throws DatabaseException {
		Query query = new Query(Type.INSERT, "TRACKS");
		String id = UUID.randomUUID().toString();
		query.addValueField(new ValueField("ID", new StringParameter(id)));
		query.addValueField(new ValueField("NAME", new StringParameter(name)));
		database.insert(query);
		return getTrackById(id);
	}

	/**
	 * Saves the given track. This also saves all relationships of the track.
	 *
	 * @param track
	 *            The track to save
	 * @throws DatabaseException
	 *             if a database error occurs
	 */
	public void saveTrack(Track track) throws DatabaseException {
		Query query = new Query(Type.UPDATE, "TRACKS");
		query.addValueField(new ValueField("TRACKS.NAME", new StringParameter(track.getName())));
		query.addValueField(new ValueField("TRACKS.REMIX", new StringParameter(track.getRemix())));
		query.addWhereClause(new ValueFieldWhereClause(new ValueField("TRACKS.ID", new StringParameter(track.getId()))));
		database.update(query);
		/* store artist information. */
		track.getArtists(); /* prefetch artists. */
		query = new Query(Type.DELETE, "TRACK_ARTISTS");
		query.addWhereClause(new ValueFieldWhereClause(new ValueField("TRACK_ARTISTS.TRACK", new StringParameter(track.getId()))));
		database.update(query);
		for (int index = 0; index < track.getArtists().size(); ++index) {
			query = new Query(Type.INSERT, "TRACK_ARTISTS");
			query.addValueField(new ValueField("TRACK_ARTISTS.TRACK", new StringParameter(track.getId())));
			query.addValueField(new ValueField("TRACK_ARTISTS.ARTIST", new StringParameter(track.getArtists().get(index).getId())));
			query.addValueField(new ValueField("TRACK_ARTISTS.DISPLAY_ORDER", new IntegerParameter(index + 1)));
			database.insert(query);
		}

		/* store party links. */
		Collection<Party> parties = track.getParties(); /* prefetch parties. */
		query = new Query(Type.DELETE, "PARTY_TRACKS");
		query.addWhereClause(new ValueFieldWhereClause(new ValueField("PARTY_TRACKS.TRACK", new StringParameter(track.getId()))));
		database.update(query);
		for (Party party : parties) {
			query = new Query(Type.INSERT, "PARTY_TRACKS");
			query.addValueField(new ValueField("PARTY_TRACKS.TRACK", new StringParameter(track.getId())));
			query.addValueField(new ValueField("PARTY_TRACKS.PARTY", new StringParameter(party.getId())));
			database.insert(query);
		}

		/* store properties. */
		saveProperties(track.getProperties(), "TRACK_PROPERTIES", "TRACK", track.getId());
	}

	/**
	 * Returns the derivative with the given ID.
	 *
	 * @param id
	 *            The ID of the derivatives to load
	 * @return The derivative with the given ID
	 * @throws DatabaseException
	 *             if a database error occurs
	 */
	public TrackDerivative getTrackDerivativeById(String id) throws DatabaseException {
		Query query = new Query(Type.SELECT, "TRACK_DERIVATIVES");
		query.addField(new Field("TRACK_DERIVATIVES.*"));
		query.addWhereClause(new ValueFieldWhereClause(new ValueField("TRACK_DERIVATIVES.ID", new StringParameter(id))));
		return loadTrackDerivativeProperties(database.getSingle(query, trackDerivativeCreator));
	}

	/**
	 * Returns the derivatives for the given track.
	 *
	 * @param trackId
	 *            The track ID to get the derivatives for
	 * @return The derivatives for the given track
	 * @throws DatabaseException
	 *             if a database error occurs
	 */
	public Collection<TrackDerivative> getTrackDerivativesByTrack(String trackId) throws DatabaseException {
		Query query = new Query(Type.SELECT, "TRACK_DERIVATIVES");
		query.addField(new Field("TRACK_DERIVATIVES.*"));
		query.addWhereClause(new ValueFieldWhereClause(new ValueField("TRACK_DERIVATIVES.TRACK", new StringParameter(trackId))));
		return loadTrackDerivativeProperties(database.getMultiple(query, trackDerivativeCreator));
	}

	/**
	 * Creates a new derivative for the given track.
	 *
	 * @param track
	 *            The track to create the derivative for
	 * @return The created derivative
	 * @throws DatabaseException
	 *             if a database error occurs
	 */
	public TrackDerivative createTrackDerivative(Track track) throws DatabaseException {
		Query query = new Query(Type.INSERT, "TRACK_DERIVATIVES");
		String id = UUID.randomUUID().toString();
		query.addValueField(new ValueField("TRACK_DERIVATIVES.ID", new StringParameter(id)));
		query.addValueField(new ValueField("TRACK_DERIVATIVES.TRACK", new StringParameter(track.getId())));
		database.insert(query);
		return getTrackDerivativeById(id);
	}

	/**
	 * Loads the properties for the given track derivative.
	 *
	 * @param trackDerivative
	 *            The track derivative to load the properties for
	 * @return The track derivative with its properties loaded
	 * @throws DatabaseException
	 *             if a database error occurs
	 */
	public TrackDerivative loadTrackDerivativeProperties(TrackDerivative trackDerivative) throws DatabaseException {
		return loadProperties(trackDerivative, "TRACK_DERIVATIVE_PROPERTIES", "TRACK_DERIVATIVE");
	}

	/**
	 * Loads the properties for the given track derivatives.
	 *
	 * @param trackDerivatives
	 *            The track derivatives to load the properties for
	 * @return The track derivatives with their properties loaded
	 * @throws DatabaseException
	 *             if a database error occurs
	 */
	public List<TrackDerivative> loadTrackDerivativeProperties(List<TrackDerivative> trackDerivatives) throws DatabaseException {
		for (TrackDerivative trackDerivative : trackDerivatives) {
			loadTrackDerivativeProperties(trackDerivative);
		}
		return trackDerivatives;
	}

	/**
	 * Saves the given track derivative. As a track derivative does not have any
	 * attributes of its own only its properties are saved.
	 *
	 * @param trackDerivative
	 *            The track derivative to save
	 * @throws DatabaseException
	 *             if a database error occurs
	 */
	public void saveTrackDerivate(TrackDerivative trackDerivative) throws DatabaseException {
		saveProperties(trackDerivative.getProperties(), "TRACK_DERIVATIVE_PROPERTIES", "TRACK_DERIVATIVE", trackDerivative.getId());
	}

	/**
	 * Removes the given track derivative and all its properties from the
	 * database.
	 *
	 * @param trackDerivative
	 *            The track derivative to remove
	 * @throws DatabaseException
	 *             if a database error occurs
	 */
	public void removeTrackDerivative(TrackDerivative trackDerivative) throws DatabaseException {
		Query query = new Query(Type.DELETE, "TRACK_DERIVATIVES");
		query.addWhereClause(new ValueFieldWhereClause(new ValueField("TRACK_DERIVATIVES.ID", new StringParameter(trackDerivative.getId()))));
		database.update(query);
		/* remove the properties. */
		saveProperties(new Properties().set("dirty", "true").removeAll(), "TRACK_DERIVATIVE_PROPERTIES", "TRACK_DERIVATIVE", trackDerivative.getId());
	}

	/**
	 * Returns all groups the artist with the given ID belongs to.
	 *
	 * @param artistId
	 *            The ID of the artist
	 * @return All groups the artist belongs to
	 * @throws DatabaseException
	 *             if a database error occurs
	 */
	public Collection<Group> getGroupsByArtist(String artistId) throws DatabaseException {
		Query query = new Query(Type.SELECT, "GROUPS");
		query.addField(new Field("GROUPS.*"));
		query.addJoin(new Join(JoinType.INNER, "GROUP_ARTISTS", new Field("GROUPS.ID"), new Field("GROUP_ARTISTS.GROUP_")));
		query.addWhereClause(new ValueFieldWhereClause(new ValueField("GROUP_ARTISTS.ARTIST", new StringParameter(artistId))));
		return database.getMultiple(query, groupCreator);
	}

	/**
	 * Returns all styles for the track with the given ID.
	 *
	 * @param trackId
	 *            The ID of the track
	 * @return All styles for the given track
	 * @throws DatabaseException
	 *             if a database error occurs
	 */
	public Collection<Style> getStylesByTrack(String trackId) throws DatabaseException {
		Query query = new Query(Type.SELECT, "STYLES");
		query.addField(new Field("STYLES.*"));
		query.addJoin(new Join(JoinType.INNER, "TRACK_STYLES", new Field("STYLES.ID"), new Field("TRACK_STYLES.STYLE")));
		query.addWhereClause(new ValueFieldWhereClause(new ValueField("TRACK_STYLES.TRACK", new StringParameter(trackId))));
		return database.getMultiple(query, styleCreator);
	}

	/**
	 * Returns all parties.
	 *
	 * @return All parties
	 * @throws DatabaseException
	 *             if a database error occurs
	 */
	public Collection<Party> getAllParties() throws DatabaseException {
		Query query = new Query(Type.SELECT, "PARTIES");
		query.addField(new Field("PARTIES.*"));
		return loadPartyProperties(database.getMultiple(query, partyCreator));
	}

	/**
	 * Returns all parties that the track with the given ID was released at.
	 *
	 * @param trackId
	 *            The ID of the track
	 * @return All parties the track was released at
	 * @throws DatabaseException
	 *             if a database error occurs
	 */
	public Collection<Party> getPartiesByTrackId(String trackId) throws DatabaseException {
		Query query = new Query(Type.SELECT, "PARTIES");
		query.addField(new Field("PARTIES.*"));
		query.addJoin(new Join(JoinType.INNER, "PARTY_TRACKS", new Field("PARTY_TRACKS.PARTY"), new Field("PARTIES.ID")));
		query.addWhereClause(new ValueFieldWhereClause(new ValueField("PARTY_TRACKS.TRACK", new StringParameter(trackId))));
		return loadPartyProperties(database.getMultiple(query, partyCreator));
	}

	/**
	 * Returns the party with the given ID.
	 *
	 * @param partyId
	 *            The ID of the party
	 * @return The party with the given ID
	 * @throws DatabaseException
	 *             if a database error occurs
	 */
	public Party getPartyById(String partyId) throws DatabaseException {
		Query query = new Query(Type.SELECT, "PARTIES");
		query.addField(new Field("PARTIES.*"));
		query.addWhereClause(new ValueFieldWhereClause(new ValueField("PARTIES.ID", new StringParameter(partyId))));
		return loadPartyProperties(database.getSingle(query, partyCreator));
	}

	/**
	 * Loads the properties of the given party.
	 *
	 * @param party
	 *            The party to load the properties for
	 * @return The party with its properties loaded
	 * @throws DatabaseException
	 *             if a database error occurs
	 */
	public Party loadPartyProperties(Party party) throws DatabaseException {
		return loadProperties(party, "PARTY_PROPERTIES", "PARTY");
	}

	/**
	 * Loads the properties of the given parties.
	 *
	 * @param parties
	 *            The parties to load the properties for
	 * @return The parties with their properties loaded
	 * @throws DatabaseException
	 *             if a database error occurs
	 */
	public List<Party> loadPartyProperties(List<Party> parties) throws DatabaseException {
		for (Party party : parties) {
			loadPartyProperties(party);
		}
		return parties;
	}

	/**
	 * Saves the given party.
	 *
	 * @param party
	 *            The party to save
	 * @throws DatabaseException
	 *             if a database error occurs
	 */
	public void saveParty(Party party) throws DatabaseException {
		Query query = new Query(Type.UPDATE, "PARTIES");
		query.addValueField(new ValueField("NAME", new StringParameter(party.getName())));
		query.addWhereClause(new ValueFieldWhereClause(new ValueField("ID", new StringParameter(party.getId()))));
		database.update(query);
		savePartyProperties(party);
	}

	/**
	 * Saves the properties of the given party.
	 *
	 * @param party
	 *            The party whose properties to save
	 * @throws DatabaseException
	 *             if a database error occurs
	 */
	public void savePartyProperties(Party party) throws DatabaseException {
		saveProperties(party.getProperties(), "PARTY_PROPERTIES", "PARTY", party.getId());
	}

	/**
	 * Creates a new party with the given name.
	 *
	 * @param name
	 *            The name of the party
	 * @return The new party
	 * @throws DatabaseException
	 *             if a database error occurs
	 */
	public Party createParty(String name) throws DatabaseException {
		Query query = new Query(Type.INSERT, "PARTIES");
		String id = UUID.randomUUID().toString();
		query.addValueField(new ValueField("ID", new StringParameter(id)));
		query.addValueField(new ValueField("NAME", new StringParameter(name)));
		database.insert(query);
		return getPartyById(id);
	}

	/**
	 * Returns the user with the given name.
	 *
	 * @param username
	 *            The name of the user
	 * @return The user, or {@code null} if the user does not exist
	 * @throws DatabaseException
	 *             if a database error occurs
	 */
	public User getUserByName(String username) throws DatabaseException {
		Query query = new Query(Type.SELECT, "USERS");
		query.addField(new Field("USERS.*"));
		query.addWhereClause(new ValueFieldWhereClause(new ValueField("USERS.NAME", new StringParameter(username))));
		return database.getSingle(query, userCreator);
	}

	/**
	 * Returns the user connected with the given OpenID.
	 *
	 * @param openId
	 *            The OpenID to find the user for
	 * @return The user connected with the given OpenID, or {@code null} if
	 *         there is no such user
	 * @throws DatabaseException
	 *             if a database error occurs
	 */
	public User getUserByOpenId(String openId) throws DatabaseException {
		Query query = new Query(Type.SELECT, "USERS");
		query.addField(new Field("USERS.*"));
		query.addJoin(new Join(JoinType.INNER, "USER_OPENIDS", new Field("USER_OPENIDS.USER"), new Field("USERS.ID")));
		query.addWhereClause(new ValueFieldWhereClause(new ValueField("USER_OPENIDS.OPENID", new StringParameter(openId))));
		return database.getSingle(query, userCreator);
	}

	/**
	 * Returns all OpenIDs connected with the user with the given ID.
	 *
	 * @param userId
	 *            The ID of the user
	 * @return All OpenIDs connected with the given user
	 * @throws DatabaseException
	 *             if a database error occurs
	 */
	public Collection<String> getOpenIdsByUser(String userId) throws DatabaseException {
		Query query = new Query(Type.SELECT, "USER_OPENIDS");
		query.addField(new Field("USER_OPENIDS.*"));
		query.addWhereClause(new ValueFieldWhereClause(new ValueField("USER_OPENIDS.USER", new StringParameter(userId))));
		return database.getMultiple(query, new StringCreator("USER_OPENIDS.OPENID"));
	}

	//
	// PRIVATE METHODS
	//

	/**
	 * Loads the properties for the given object.
	 *
	 * @param <T>
	 *            The type of the object
	 * @param object
	 *            The object
	 * @param table
	 *            The table to load the properties from
	 * @param type
	 *            The type of the object (“ARTIST,” “TRACK,” etc.)
	 * @return The object with its properties loaded
	 * @throws DatabaseException
	 *             if a database error occurs
	 */
	private <T extends Base> T loadProperties(final T object, final String table, String type) throws DatabaseException {
		if (object == null) {
			return null;
		}
		Query query = new Query(Type.SELECT, table);
		query.addField(new Field(table + ".PROPERTY"));
		query.addField(new Field(table + ".VALUE"));
		query.addWhereClause(new ValueFieldWhereClause(new ValueField(type, new StringParameter(object.getId()))));
		database.process(query, new ResultProcessor() {

			@Override
			public void processResult(ResultSet resultSet) throws SQLException {
				if (resultSet.isFirst()) {
					object.getProperties().removeAll();
				}
				object.getProperties().set(resultSet.getString(table + ".PROPERTY"), resultSet.getString(table + ".VALUE"));
			}

		});
		return object;
	}

	/**
	 * {@link Artist} implementation that retrieves some attributes (such as
	 * {@link #getGroups()}, and {@link #getTracks()}) from the
	 * {@link DataManager} on demand.
	 *
	 * @author <a href="mailto:bombe@pterodactylus.net">David ‘Bombe’ Roden</a>
	 */
	private class LazyArtist extends DefaultArtist {

		/** Memoizer for the tracks by this artist. */
		private final Memoizer<Void> tracksMemoizer = new Memoizer<Void>(new Callable<Void>() {

			@Override
			public Void call() throws DatabaseException {
				if (!hasValue("tracks")) {
					getValue("tracks", Collection.class).set(getTracksByArtist(getId()));
				}
				return null;
			}
		});

		/** Memoizer for the groups of this artist. */
		private final Memoizer<Void> groupsMemoizer = new Memoizer<Void>(new Callable<Void>() {

			@Override
			public Void call() throws Exception {
				if (!hasValue("groups")) {
					getValue("groups", Collection.class).set(getGroupsByArtist(getId()));
				}
				return null;
			}

		});

		/**
		 * Creates a new lazy artist.
		 *
		 * @param id
		 *            The ID of the artist
		 */
		public LazyArtist(String id) {
			super(id);
		}

		//
		// DEFAULTARTIST METHODS
		//

		/**
		 * {@inheritDoc}
		 */
		@Override
		public Collection<Group> getGroups() {
			groupsMemoizer.get();
			return super.getGroups();
		}

		/**
		 * {@inheritDoc}
		 */
		@Override
		public Collection<Track> getTracks() {
			tracksMemoizer.get();
			return super.getTracks();
		}

	}

	/**
	 * {@link ObjectCreator} implementation that can create {@link Artist}
	 * objects. This specific class actually creates {@link LazyArtist}
	 * instances.
	 *
	 * @author <a href="mailto:bombe@pterodactylus.net">David ‘Bombe’ Roden</a>
	 */
	private class ArtistCreator implements ObjectCreator<Artist> {

		/**
		 * {@inheritDoc}
		 */
		@Override
		public Artist createObject(ResultSet resultSet) throws SQLException {
			return new LazyArtist(resultSet.getString("ARTISTS.ID")).setName(resultSet.getString("ARTISTS.NAME"));
		}

	}

	/**
	 * {@link Group} implementation that retrieves some attributes (such as
	 * {@link #getArtists()}) from the {@link DataManager} on demand.
	 *
	 * @author <a href="mailto:bombe@pterodactylus.net">David ‘Bombe’ Roden</a>
	 */
	private class LazyGroup extends DefaultGroup {

		/** Memoizer for the artist. */
		private final Memoizer<Void> artistsMemoizer = new Memoizer<Void>(new Callable<Void>() {

			@Override
			public Void call() throws Exception {
				if (!hasValue("artists")) {
					getValue("artists", Collection.class).set(getArtistsByGroup(getId()));
				}
				return null;
			}

		});

		/**
		 * Creates a new lazy group.
		 *
		 * @param id
		 *            The ID of the group
		 */
		public LazyGroup(String id) {
			super(id);
		}

		//
		// DEFAULTGROUP METHODS
		//

		/**
		 * {@inheritDoc}
		 */
		@Override
		public Collection<Artist> getArtists() {
			artistsMemoizer.get();
			return super.getArtists();
		}

	}

	/**
	 * {@link ObjectCreator} implementation that can create {@link Group}
	 * objects. This specific implementation creates {@link LazyGroup}
	 * instances.
	 *
	 * @author <a href="mailto:bombe@pterodactylus.net">David ‘Bombe’ Roden</a>
	 */
	private class GroupCreator implements ObjectCreator<Group> {

		/**
		 * {@inheritDoc}
		 */
		@Override
		public Group createObject(ResultSet resultSet) throws SQLException {
			return new LazyGroup(resultSet.getString("GROUPS.ID")).setName(resultSet.getString("GROUPS.NAME")).setUrl(resultSet.getString("GROUPS.URL"));
		}

	}

	/**
	 * {@link Track} implementation that retrieves some attributes (such as
	 * {@link #getArtists()}, and {@link #getStyles()}) from the
	 * {@link DataManager} on demand.
	 *
	 * @author <a href="mailto:bombe@pterodactylus.net">David ‘Bombe’ Roden</a>
	 */
	private class LazyTrack extends DefaultTrack {

		/** Memoizer for the artists. */
		private final Memoizer<Void> artistsMemoizer = new Memoizer<Void>(new Callable<Void>() {

			@Override
			public Void call() throws Exception {
				if (!hasValue("artists")) {
					getValue("artists", List.class).set(getArtistsByTrack(getId()));
				}
				return null;
			}

		});

		/** Memoizer for the styles. */
		private final Memoizer<Void> stylesMemoizer = new Memoizer<Void>(new Callable<Void>() {

			@Override
			public Void call() throws Exception {
				if (!hasValue("styles")) {
					getValue("styles", Collection.class).set(getStylesByTrack(getId()));
				}
				return null;
			}

		});

		/** Memoizer for the remix artists. */
		private final Memoizer<Void> remixArtistsMemoizer = new Memoizer<Void>(new Callable<Void>() {

			@Override
			public Void call() throws Exception {
				if (!hasValue("remixArtists")) {
					getValue("remixArtists", List.class).set(getRemixArtistsByTrack(getId()));
				}
				return null;
			}

		});

		/** Memoizer for the track derivatives. */
		private final Memoizer<Void> derivativesMemoizer = new Memoizer<Void>(new Callable<Void>() {

			@Override
			public Void call() throws Exception {
				if (!hasValue("derivatives")) {
					getValue("derivatives", Collection.class).set(getTrackDerivativesByTrack(getId()));
				}
				return null;
			}

		});

		/** Memoizer for the related tracks. */
		private final Memoizer<Void> relatedTracksMemoizer = new Memoizer<Void>(new Callable<Void>() {

			@Override
			public Void call() throws Exception {
				if (!hasValue("relatedTracks")) {
					getValue("relatedTracks", Map.class).set(getRelatedTracksByTrack(getId()));
				}
				return null;
			}

		});

		/**
		 * Creates a new track.
		 *
		 * @param id
		 *            The ID of the track
		 */
		public LazyTrack(String id) {
			super(id);
		}

		//
		// DEFAULTTRACK METHODS
		//

		/**
		 * {@inheritDoc}
		 */
		@Override
		public List<Artist> getArtists() {
			artistsMemoizer.get();
			return super.getArtists();
		}

		/**
		 * {@inheritDoc}
		 */
		@Override
		public Collection<Style> getStyles() {
			stylesMemoizer.get();
			return super.getStyles();
		}

		/**
		 * {@inheritDoc}
		 */
		@Override
		public List<Artist> getRemixArtists() {
			remixArtistsMemoizer.get();
			return super.getRemixArtists();
		}

		/**
		 * {@inheritDoc}
		 */
		@Override
		public Collection<TrackDerivative> getDerivatives() {
			derivativesMemoizer.get();
			return super.getDerivatives();
		}

		/**
		 * {@inheritDoc}
		 */
		@Override
		public Map<Relationship, Collection<Track>> getRelatedTracks() {
			relatedTracksMemoizer.get();
			return super.getRelatedTracks();
		}

		/**
		 * {@inheritDoc}
		 */
		@Override
		public Collection<Party> getParties() {
			if (!hasValue("parties")) {
				try {
					getValue("parties", Collection.class).set(getPartiesByTrackId(getId()));
				} catch (DatabaseException de1) {
					throw new RuntimeException("Could not load Parties for Track " + getId() + ".", de1);
				}
			}
			return super.getParties();
		}

	}

	/**
	 * {@link ObjectCreator} implementation that can create {@link Track}
	 * objects. This specific implementation creates {@link LazyTrack}
	 * instances.
	 *
	 * @author <a href="mailto:bombe@pterodactylus.net">David ‘Bombe’ Roden</a>
	 */
	private class TrackCreator implements ObjectCreator<Track> {

		/**
		 * {@inheritDoc}
		 */
		@Override
		public Track createObject(ResultSet resultSet) throws SQLException {
			return new LazyTrack(resultSet.getString("TRACKS.ID")).setName(resultSet.getString("TRACKS.NAME")).setRemix(resultSet.getString("TRACKS.REMIX"));
		}

	}

	/**
	 * {@link ObjectCreator} implementation that can create
	 * {@link TrackDerivative} objects.
	 *
	 * @author <a href="mailto:bombe@pterodactylus.net">David ‘Bombe’ Roden</a>
	 */
	private class TrackDerivativeCreator implements ObjectCreator<TrackDerivative> {

		/**
		 * {@inheritDoc}
		 */
		@Override
		public TrackDerivative createObject(ResultSet resultSet) throws SQLException {
			return new DefaultTrackDerivative(resultSet.getString("TRACK_DERIVATIVES.ID"));
		}

	}

	/**
	 * {@link ObjectCreator} implementation that can create {@link Style}
	 * objects.
	 *
	 * @author <a href="mailto:bombe@pterodactylus.net">David ‘Bombe’ Roden</a>
	 */
	private class StyleCreator implements ObjectCreator<Style> {

		/**
		 * {@inheritDoc}
		 */
		@Override
		public Style createObject(ResultSet resultSet) throws SQLException {
			return new DefaultStyle(resultSet.getString("STYLES.ID")).setName(resultSet.getString("STYLES.NAME"));
		}

	}

	/**
	 * {@link Party} implementation that loads additional information only on
	 * demand.
	 *
	 * @author <a href="mailto:bombe@pterodactylus.net">David ‘Bombe’ Roden</a>
	 */
	private class LazyParty extends DefaultParty {

		/**
		 * Creates a new party.
		 *
		 * @param id
		 *            The ID of the party
		 */
		public LazyParty(String id) {
			super(id);
		}

		//
		// PARTY METHODS
		//

		/**
		 * {@inheritDoc}
		 */
		@Override
		public Collection<Track> getReleases() {
			if (!hasValue("releases")) {
				try {
					getValue("releases", Collection.class).set(getTracksByParty(getId()));
				} catch (DatabaseException de1) {
					throw new RuntimeException("Could not loaded Tracks for Party " + getId() + ".", de1);
				}
			}
			return super.getReleases();
		}

	}

	/**
	 * {@link ObjectCreator} implementation that can create {@link Party}
	 * objects.
	 *
	 * @author <a href="mailto:bombe@pterodactylus.net">David ‘Bombe’ Roden</a>
	 */
	private class PartyCreator implements ObjectCreator<Party> {

		/**
		 * {@inheritDoc}
		 */
		@Override
		public Party createObject(ResultSet resultSet) throws SQLException {
			return new LazyParty(resultSet.getString("PARTIES.ID")).setName(resultSet.getString("PARTIES.NAME"));
		}

	}

	/**
	 * {@link User} implementation that retrieves some attributes (such as
	 * {@link #getOpenIds()}) from the {@link DataManager} on demand.
	 *
	 * @author <a href="mailto:bombe@pterodactylus.net">David ‘Bombe’ Roden</a>
	 */
	private class LazyUser extends DefaultUser {

		/** Memoizer for a user’s OpenIDs. */
		private final Memoizer<Void> openIdMemoizer = new Memoizer<Void>(new Callable<Void>() {

			@Override
			public Void call() throws Exception {
				if (!hasValue("openIds")) {
					getValue("openIds", Collection.class).set(getOpenIdsByUser(getId()));
				}
				return null;
			}
		});

		/**
		 * Creates a new user.
		 *
		 * @param id
		 *            The ID of the user
		 */
		public LazyUser(String id) {
			super(id);
		}

		/**
		 * {@inheritDoc}
		 */
		@Override
		public Collection<String> getOpenIds() {
			openIdMemoizer.get();
			return super.getOpenIds();
		}

	}

	/**
	 * {@link ObjectCreator} implementation that can create {@link User}
	 * objects. This specific implementation creates {@link LazyUser} instances.
	 *
	 * @author <a href="mailto:bombe@pterodactylus.net">David ‘Bombe’ Roden</a>
	 */
	private class UserCreator implements ObjectCreator<User> {

		/**
		 * {@inheritDoc}
		 */
		@Override
		public User createObject(ResultSet resultSet) throws SQLException {
			return new LazyUser(resultSet.getString("USERS.ID")).setName(resultSet.getString("USERS.NAME")).setPasswordHash(resultSet.getString("USERS.PASSWORD")).setLevel(resultSet.getInt("USERS.LEVEL"));
		}

	}

}
